/*
 * Last updated: 28th dec 1993
 * This program contains serveral dither subroutines.
 * 1) Space filling curve halftoning
 * 2) Space filling curve halftoning with adaptive clustering +
 *    selective precipitation
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define TRUE    1
#define FALSE   0

#define BLACK   255
#define WHITE   0

#define LEFT    0
#define RIGHT   1
#define UP      2
#define DOWN    3
#define END     255

#define sq(x) ((x)*(x))
#define move(x,y)  switch (path[x][y])          \
                   {                            \
                     case UP:   y++; break;     \
                     case DOWN: y--; break;     \
                     case LEFT: x--; break;     \
                     case RIGHT:x++; break;     \
                   }
#define Pixel(p,x,y) ((p)[(x)/NDUP][(y)/NDUP])
#define GetMem(p,t,w,h) { int i;								\
			  if (((p)=(t **)malloc(sizeof(t *)*(w)))==NULL) 			\
  		      	  {									\
			    fprintf (stderr, "[GetMem]: not enough for **\n");			\
    			    exit (1);								\
  		          }									\
  		          for (i=0; i<(w) ; i++)						\
    			    if (((p)[i]=(t *)malloc(sizeof(t)*(h)))==NULL)			\
    		    	    {									\
			      fprintf (stderr, "[GetMem]: not enough memory for  *\n");		\
      			      exit (1);								\
    		 	    }									\
    		 	}
#define FreeMem(p,w,h)   {				\
			    int i; 			\
			    for (i=0 ; i<(w) ; i++)	\
			      free ((p)[i]);		\
			    free((p));			\
			  }
#define ErrExit(s) {                        \
		     fprintf (stderr, s);   \
		     exit (1);              \
		   }

int LEVEL=8;
int NDUP=1;
int IXMAX=256, IYMAX=256;  /* input buffer width and height */
int OXMAX=256, OYMAX=256;  /* output buffer width and height */
int **picture;             /* input buffer */
unsigned char **path;	   /* space filling curve path */
int **out;                 /* output buffer */

/* global variables only for spiral dither */
char *sdrow, *sdcol;	   /* 2 arrays used by spiral dithering */
int *sdindex;              /* for spiral dither */
long *sdsum;		   /* for spiral dither */
float *sdkey; 		   /* for spiral dither */
unsigned char **sdmask;	   /* bit mask for each pixel in a cluster */
int **sdbuffer;


/********************************************
 * generate picture intensity from 0 to 255 *
 ********************************************/
void genpicture(void)           /* generate a picture           */
{                               /* you can modify it to read in */
  int i,j,k,l,r;                /* from a file                  */

  int RADIUS= (IXMAX > IYMAX)? IYMAX/4 : IXMAX/4;

  for (i=0; i<IXMAX; i++)
    for (j=0; j<IYMAX; j++)
      if ((r=(sq(i-(IXMAX/2))+sq(j-(IYMAX/2))))>sq(RADIUS))
        picture[i][j]=((IYMAX-j-1)/(double) IYMAX)*256.0;
      else
        picture[i][j]=(int)(sqrt((double)r)/(double)RADIUS*256.0);
}


/**********************************\
*  readpicture(): read a picture  *
\**********************************/
void readpicture(char filename[])
{
  int i,j;
  FILE *fptr;
  unsigned char *buffer;
  
  if ((buffer=(unsigned char *)malloc(sizeof(unsigned char)*IXMAX))==NULL)
    ErrExit ("[readpicture]: Not enough memory!\n");
  if ((fptr=fopen(filename,"rb"))==NULL)
    ErrExit ("[readpicture]: cannot open image file\n");

  for (i=0 ; i<IYMAX ; i++)
  {
    fread(buffer, IXMAX, 1, fptr); 
    for (j=0 ; j<IXMAX; j++)
      picture[j][i] = 255-(int)buffer[j];
  }
  fclose(fptr);
}


/****************************************************\
 * outps(): output a picture as a postscript image *
\****************************************************/
void outps(int **pic, int scale, int dpi)
{
  int i,j;
  printf("%%!\n/picstr %d string def\n100 100 translate\n",OXMAX);
  printf("gsave %f %f scale %d %d 8 [%d 0 0 -%d 0 %d]\n"
         "{currentfile picstr readhexstring pop} image\n",
         OXMAX*scale*72.0/(float)dpi, OYMAX*scale*72.0/(float)dpi,
         OXMAX, OYMAX,
         OXMAX, OYMAX, OYMAX);
  for (j=0; j<OYMAX; j++)
  {
    for (i=0; i<OXMAX; i++)
      printf("%02X", 255^pic[i][j]);
    printf("\n");
  }
  printf("grestore\n");
  printf("showpage\n");
}




/*****************************************************\
 * genspacefill(): generates the space filling trace *
\*****************************************************/
static int pathx,pathy;
void step(long angle)
{
  while (angle > 270)     /* Fold ANGLE to be 0, 90, 180, 270 */
    angle -= 360;    
  while (angle <   0)
    angle += 360;
  if (angle == 0)
    path[pathx++][pathy]=RIGHT; /* +X */
  else if (angle == 90)
    path[pathx][pathy++]=UP;    /* +Y */
  else if (angle == 180)
    path[pathx--][pathy]=LEFT;  /* -X */
  else if (angle == 270)
    path[pathx][pathy--]=DOWN;  /* -Y */
}

void hilbert (long orient,long *angle, long level)
{
  long sw;

  if (level-- <= 0)
    return;
  *angle += orient * 90;
  hilbert(-orient,angle,level);
  step(*angle);
  *angle -= orient * 90;
  hilbert(orient,angle,level);
  step(*angle);
  hilbert(orient,angle,level);
  *angle -= orient * 90;
  step(*angle);
  hilbert(-orient,angle,level);
  *angle += orient * 90;
}

void genspacefill(void)
{
 int i,j;
 long angle=0;

 pathx=pathy=0;                         /* start of the trace */
 hilbert(1,&angle,LEVEL);               /* generates a hilbert curve */
 path[pathx][pathy]=(unsigned char)END; /* indicates end of path */
}



/*==========================================================================*
 * Halftoning using Original Space Filling Curve                            *
 * Limitation (10th Dec 1993)
 * Can only process image with size 2^n x 2^n where n is 1,2,3, ...
 *==========================================================================*/
void spacefill(int **picture, int **out, int clustersize)
{
  int accumulator=0, fromx=0,fromy=0, i,j,x,y;
  char Go_on=TRUE;

  genspacefill();    /* generates the spacefilling path    */
  do         /* repeats until all pixels processed */
  {
    for (x=fromx,y=fromy,i=0; i<clustersize && Go_on; i++)
    {
      int dir=path[x][y];
      accumulator+=Pixel(picture,x,y);  /* gathers the total value of the cluster */
      switch(dir)
      {
        case UP:   y++; break;
        case DOWN: y--; break;
        case LEFT: x--; break;
        case RIGHT:x++; break;
        default: Go_on=FALSE;     /* end of path ? */
      }
    } /* for */
    Go_on=TRUE;

    for (x=fromx,y=fromy,i=0; i<clustersize && Go_on; i++)
    {
      int dir=path[x][y];
      if (accumulator>=BLACK)  /* precipitates */
      {
        out[x][y]=BLACK;
        accumulator-=BLACK;
      }
      else out[x][y]=WHITE;
      switch(dir)
      {
        case UP:   y++; break;
        case DOWN: y--; break;
        case LEFT: x--; break;
        case RIGHT:x++; break;
        default: Go_on=FALSE;     /* end of path ? */
      } /* switch */
      fromx=x; fromy=y;
    }
  } while (Go_on);
}



/*==========================================================================*
 * Halftoning using Space Filling Curve with  adaptive clustering and
 * selective precipitation
 * Limitation: (10th dec 93)
 * Only process image with size 2^n x 2^n where n is +ve integer.
 *
 * 1) To switch off the selective precipitation, pass do_sp = FALSE.
 * 2) To switch off the adaptive clustering, set threshold = inifinity or do_ac=FALSE
 *==========================================================================*/
/* 
 * Description of the pointer along the space filling curve.
 * 
 * clusterx,			      windowx,	  currx,     frontx,
 * clustery			      windowy	  curry      fronty
 *    |					 |          |           |
 *    v					 v	    v		v
 *   +----------------------------------------------------------+
 *   |                        Cluster                           |
 *   +----------------------------------------------------------+
 *                                       |                      |
 *                                       |                      |
 *                                       |          /\          |
 *                                       |        /    \        |
 *                                       |___   /        \   ___|
 *                                       |    \/           \/   |
 *					Laplacian of Gaussian Filter
 */
void spacefilterwindow(int **picture, int **out, int maxclustersize, int thresh, char do_sp, char do_ac)
{
  char change;           /* Flag indicate sudden change detected */
  char ending; 		 /* flag indicates end of space filling curve */
  int accumulator;       /* Accumulate gray value */
  int currclustersize;   /* Record size of current cluster */
  int frontx, fronty;    /* Pointer to the front of the cluster */
  int windowx, windowy;  /* Pointer to first pixel applied with filter */
  int clusterx, clustery;/* Pointer to first pixel in current cluster */
  int windowlen;	 /* Size of the moving window */
  int winsum;		 /* Current moving window's sum */
  int maxsum;		 /* Maximum moving window's sum recorded */
  int start;		 /* Position of the moving window with max sum */
  int *cluster;		 /* An array hold the pixel of current cluster */
  int last, i,j, tempx, tempy, currx, curry;	/* temp variables */
  long filter[7] = {-1, -5, 0, 13, 0, -5, -1};  /* filter */ 
  long convolution;	 /* Convolution value in this turn */
  long lastconvolution;  /* Convolution value in last turn */
  long clustercount=0;	 /* Record the total number of clusters */

  if ((cluster=malloc(sizeof(int)*maxclustersize))==NULL)
  {
    fprintf(stderr,"not enough memory for cluster\n");
    return;
  }
  genspacefill();    /* generates the spacefilling path    */
  
  convolution=0;
  currclustersize=0;
  accumulator=0;
  for (frontx=0, fronty=0, i=0 ; i<7 ; i++)
  {
    if (i<3)
    {
      cluster[currclustersize] = Pixel(picture,frontx,fronty);
      accumulator += cluster[currclustersize];
      currclustersize++;
    }
    if (i==3)
    {  currx = frontx;  curry = fronty;  }
    convolution += filter[i]*(long)(Pixel(picture,frontx,fronty));
    move(frontx,fronty);  /* assume the image at least has 7 pixels */
  }
  lastconvolution = convolution;
  clusterx=0;   clustery=0;
  windowx=0;    windowy=0;
  change=FALSE;
  ending=FALSE;

  while (TRUE)
  {
    if (do_ac) /* switch on/off adaptive clustering */
    {
      /* do convolution */
      convolution = 0;
      for (tempx=windowx, tempy=windowy, i=0 ; i<7 ; i++)
      {
        convolution += filter[i]*Pixel(picture,tempx,tempy);
        move(tempx,tempy);
      }

      /* detect sudden change */
      if ( (convolution >= 0 && lastconvolution <=0 
            && abs(convolution-lastconvolution)>thresh)
         ||(convolution <= 0  && lastconvolution >=0 
            && abs(convolution-lastconvolution)>thresh)) 
        change=TRUE; /* force output pixel */
    }    

    /* Output pixels if necessary */
    if (change || currclustersize >= maxclustersize || ending) 
    {
      change=FALSE;
      clustercount++;
      
      /* Search the best position withn cluster to preciptate */
      start = 0;
      if (do_sp) /* switch on/off selective precipitation */
      {
        windowlen = accumulator/BLACK;
        winsum = 0;
        for (i=0; i<windowlen; i++)
          winsum += cluster[i];
        for (maxsum=winsum, last=0; i<currclustersize; i++, last++)
        {
          winsum+= cluster[i] - cluster[last];
          if (winsum > maxsum)
          {
            start=last+1;
            maxsum=winsum;
          }
        }
      }

      /* Output pixels */
      for (i=0 ; currclustersize!=0 ; currclustersize--, i++)
      {
        if (accumulator>=BLACK && i>=start)  /* precipitates */
        {
          out[clusterx][clustery]=BLACK;
          accumulator-=BLACK;
        }
        else
          out[clusterx][clustery]=WHITE;
        move(clusterx,clustery)
      } /* for */

      if (ending)
	break;
    } /* if */

    cluster[currclustersize] = Pixel(picture,currx,curry);
    accumulator += cluster[currclustersize];
    currclustersize++;
    if (path[currx][curry]==END)
      ending = TRUE;
    move(currx,curry);
    move(windowx,windowy);
    move(frontx,fronty);
  } /* while */
  fprintf (stderr, "no of cluster = %li\naverage cluster size = %f\n",
           clustercount, (float)OXMAX*OYMAX/clustercount);
}




/************************************\
 * main(): output to stdout         *
\************************************/
#define  POSTSCRIPT  1
#define  HIPS	     2
int main(int argc, char *argv[])
{
  int clustersize=9, scale=2, threshold=100, i, dpi=600;
  int clusterwidth=4, clusterheight=4;
  char inputfile[256], noinput=TRUE, outformat=HIPS, do_sp=TRUE, do_ac=TRUE;
  char nosnake=FALSE, do_dd=FALSE;
    
  /* init global variables */
  IXMAX = 256;
  IYMAX = 256;
  NDUP = 1;
  OXMAX = 256;
  OYMAX = 256;
  LEVEL = 8;

  for (i=1 ; i<argc ; i++)
    if (argv[i][0] == '-')
      switch (argv[i][1])
      {
        case 'c' : i++;
	  	   if (i<argc)
	 	   {
 		     clustersize = atoi(argv[i]);
	 	     if (clustersize <= 0)
	 	       clustersize = 1;
	 	   }
 		   break;

	case 'D' : i++;
                   if (i<argc)
		   {
		     int n;
		     n = atoi(argv[i]);
		     if (n<=0 || n>3)
		       ErrExit ("Hey what are you giving me!");
		     NDUP = (int)pow((double)2, (double)(n-1));
		   }
	           break;

        case 'd' : i++;
                   if (i<argc)
                   {
                     dpi = atoi(argv[i]);
                     if (dpi <= 0)
                       dpi = 0;
                   }
                   break;
                       
        case 'i' : IXMAX = 256;
                   IYMAX = 256;
        	   i++;
 	 	   if (i<argc)
 		   {
 		     noinput = FALSE;
 		     strcpy(inputfile, argv[i]);
 		   }
 		   break;

 	case 'p' : outformat = POSTSCRIPT;
 		   break;

 	case 's' : i++;
 		   if (i<argc)
 		   {
 		     scale = atoi(argv[i]);
 		     if (scale <= 0)
 		       scale = 1;
 		   }
 		   break;

 	case 't' : i++;
 		   if (i<argc)
 		   {
 		     threshold = atoi(argv[i]);
 		     if (threshold < 0)
 		       threshold = 0;
 		   }
 		   break;

        case '-' : if (argv[i][2]=='s' && argv[i][3]=='p')       /* --sp */
                     do_sp = FALSE;
                   else if (argv[i][2]=='a' && argv[i][3]=='c')  /* --ac */
                     do_ac = FALSE;
                   else if (argv[i][2]=='n' && argv[i][3]=='s')  /* --ns */
                     nosnake = TRUE;
                   else if (argv[i][2]=='d' && argv[i][3]=='d')  /* --dd */
                     do_dd = TRUE;  
	  	   break;   
        
	default: fprintf (stderr,"Unrecognized switch %s\n", argv[i]);
      }
    
  /* allocate memory for input and output image */
  OXMAX = IXMAX * NDUP;
  OYMAX = IXMAX * NDUP;
  LEVEL = (OXMAX>OYMAX)? (int)ceil(log2((double)OXMAX)) : (int)ceil(log2((double)OYMAX));
  fprintf (stderr, "LEVEL=%d\n", LEVEL); 
  GetMem(picture, int, IXMAX, IYMAX);
  GetMem(out, int, OXMAX, OYMAX);

  /* start dither */
  if (!noinput)
    readpicture(inputfile);
  else 
    genpicture();

  if (strcmp(argv[0],"sfill")==0)
  {
    GetMem(path, unsigned char, OXMAX, OYMAX);
    spacefill(picture, out, clustersize);
    FreeMem(path,OXMAX,OYMAX);
  }  
  else if (strcmp(argv[0], "sfwindow")==0)
  {
    GetMem(path, unsigned char, OXMAX, OYMAX);
    spacefilterwindow(picture, out, clustersize, threshold, do_sp, do_ac);
    FreeMem(path,OXMAX,OYMAX);
  }
  else
    ErrExit ("Unknown command\n");

  if (outformat == POSTSCRIPT)
    outps(out, scale, dpi);

  FreeMem (picture,IXMAX,IYMAX);
  FreeMem (out,OXMAX,OYMAX);
  return 0;
}



